import os.path
from itertools import count
from types import GeneratorType
from typing import Union

from ovld import ovld

from .textgen import Context

# CSS for hrepr
styledir = f"{os.path.dirname(__file__)}/style"
css_nbreset = open(f"{styledir}/nbreset.css", encoding="utf-8").read()

# Used by __str__, set by __init__ to avoid a circular dependency
standard_html = None

current_autoid = count()


def _nextid():
    return f"AID_{next(current_autoid)}"


def flatten(seq):
    results = []
    for element in seq:
        if isinstance(element, (list, tuple)):
            results.extend(flatten(element))
        else:
            results.append(element)
    return results


class Tag:
    """
    Representation of an HTML tag.

    .. note::

        Create tags using an instance of the HTML class.

    A Tag should be immutable, but you can create modified versions
    of a tag by calling it with additional attributes or children,
    or indexing it with class names:

    >>> H = HTML()
    >>> tag = H.div()
    >>> tag
    <div></div>
    >>> tag(1, 2, 3)
    <div>123</div>
    >>> tag['klass']
    <div class="klass"></div>
    >>> tag(id='dividi')
    <div id="dividi"></div>
    >>> tag['klass'](id='dividi')(1, 2, 3)
    <div class="klass" id="dividi">123</div>

    Underscores in attributes are converted to dashes when they are
    given as keyword arguments. Attribute names with underscores can
    be provided via the ``fill`` method.

    Attributes:
        name (str): The tag name (div, span, a, etc.)
        attributes (dict): A dictionary of attributes. Attributes that
            start with ``__`` or ``--`` are reserved for internal use.
        children (tuple): Children of this node.
        resources (tuple): List of resources needed by this node or its
            children.
    """

    specialized_tags = {}

    @classmethod
    def specialize(cls, name):
        """Return a new subclass specialized for the given tag name."""
        assert cls is Tag
        if name not in cls.specialized_tags:
            cls.specialized_tags[name] = type(f"Tag::{name}", (Tag,), {})
        return cls.specialized_tags[name]

    def __init__(self, name, attributes=None, children=None, resources=None):
        self.name = name
        self.attributes = attributes or {}
        self.children = tuple(flatten(children or ()))
        self.resources = () if resources is None else tuple(resources)

    def fill(self, children=None, attributes=None, resources=None):
        if not children and not attributes and not resources:
            return self
        children = (*self.children, *children) if children else self.children
        attributes = (
            {**self.attributes, **attributes} if attributes else self.attributes
        )
        if isinstance(resources, Tag):
            resources = (resources,)
        resources = (
            tuple([*self.resources, *resources])
            if resources
            else self.resources
        )
        return type(self)(
            name=self.name,
            attributes=attributes,
            children=children,
            resources=resources,
        )

    def get_attribute(self, attr, dflt):
        return self.attributes.get(attr, dflt)

    def parts_and_resources(self):
        return standard_html.generate(self)

    def text_parts(self):
        return self.parts_and_resources()[0]

    def pretty(self, **config):
        rval, _ = self.text_parts().format(
            Context(
                tabsize=4,
                max_col=80,
                offset=0,
                line_offset=0,
                overflow="allow",
            ).replace(**config)
        )
        return rval

    def autoid(self):
        return self(id=_nextid())

    def __getitem__(self, items):
        if not isinstance(items, tuple):
            items = (items,)
        assert all(isinstance(item, str) for item in items)
        attributes = {}
        classes = [it for it in items if not it.startswith("#")]
        if classes:
            attributes["class"] = [*self.attributes.get("class", ()), *classes]
        ids = [it for it in items if it.startswith("#")]
        if ids:
            the_id = ids[-1][1:]
            if the_id == "":
                the_id = f"AID_{next(current_autoid)}"
            attributes["id"] = the_id
        return self.fill(attributes=attributes)

    def __call__(self, *children, **attributes):
        if "__constructor" in attributes:
            if "id" not in attributes and "id" not in self.attributes:
                attributes["id"] = _nextid()
        attributes = {
            attr.replace("_", "-"): value for attr, value in attributes.items()
        }
        if len(children) > 0 and isinstance(children[0], dict):
            attributes = {**children[0], **attributes}
            children = children[1:]
        return self.fill(children, attributes)

    def __eq__(self, other):
        return (
            isinstance(other, Tag)
            and self.name == other.name
            and self.attributes == other.attributes
            and self.children == other.children
            and self.resources == other.resources
        )

    def __hash__(self):
        return hash(
            (
                self.name,
                tuple(self.attributes.items()),
                self.children,
                self.resources,
            )
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.pretty(max_col=None)

    def _repr_html_(self):  # pragma: no cover
        """
        Jupyter Notebook hook to print this element as HTML.
        """
        body, extra, resources = self.parts_and_resources()
        elem = H.div(
            H.style(css_nbreset), resources, H.div["hrepr"](body, extra)
        )
        return str(elem)

    def as_page(self, pretty=False, **format_options):
        """
        Wrap this Tag as a self-contained webpage. Create a page with
        the following structure:

        .. code-block:: html

            <!DOCTYPE html>
            <html>
              <head>
                <meta http-equiv="Content-type"
                      content="text/html"
                      charset="UTF-8" />
                {self.resources}
              </head>
              <body>
                {self}
              </body>
            </html>
        """
        body, extra, resources = self.parts_and_resources()
        H = HTML()
        utf8 = H.meta(
            {"http-equiv": "Content-type"}, content="text/html", charset="UTF-8"
        )
        node = H.inline(
            H.raw("<!DOCTYPE html>"),
            H.html(H.head(utf8, resources), H.body(body, extra)),
        )
        if not pretty:
            format_options["max_col"] = None
        return node.pretty(**format_options)


@ovld
def iterate_children(self, children: Union[list, tuple, GeneratorType]):
    for child in children:
        yield from self(child)


@ovld
def iterate_children(self, child: object):
    yield child


class HTML:
    """
    Tag factory:

    >>> H = HTML()
    >>> H.div()
    <div></div>
    >>> H.a(href="http://cool.cool.cool")("My cool site")
    <a href="http://cool.cool.cool">My cool site</a>
    >>> H.span["cool", "beans"]("How stylish!")
    <span class="cool beans">How stylish!</span>
    """

    def __init__(self, tag_class=Tag, instantiate=True):
        self.tag_class = tag_class
        self.instantiate = instantiate

    def __getattr__(self, tag_name):
        tag_name = tag_name.replace("_", "-")
        tag_class = self.tag_class
        if hasattr(tag_class, "specialize"):
            tag_class = tag_class.specialize(tag_name)
        if self.instantiate:
            return tag_class(tag_name)
        else:
            return tag_class


H = HTML(tag_class=Tag, instantiate=True)
HType = HTML(tag_class=Tag, instantiate=False)



import json
from typing import Union

from ovld import ovld

from .h import HType, Tag
from .resource import JSExpression, Resource
from .textgen import Breakable, Sequence, Text, join

_type = type


##################################
# Embedding data in an attribute #
##################################


@ovld
def attr_embed(self, attr: str, value: bool):
    if value is False:
        return None
    elif value is True:
        return value


@ovld
def attr_embed(self, attr: str, value: Union[str, int, float]):
    return str(value)


@ovld
def attr_embed(self, attr: str, value: Union[list, tuple, set, frozenset]):
    return " ".join(self(attr, cls) for cls in value)


@ovld
def attr_embed(self, attr: str, expr: JSExpression):
    return expr


@ovld
def attr_embed(self, attr: str, t: Tag):
    tag_id = t.get_attribute("id", None)
    if not tag_id:
        raise ValueError(f"Cannot embed <{t.name}> element without an id.")
    return f"#{tag_id}"


@ovld
def attr_embed(self, attr: str, value: object):
    raise TypeError(
        f"Resources of type {type(value).__name__} cannot be serialized as an attribute."
    )


################################
# Embedding data in JavaScript #
################################


@ovld
def js_embed(self, d: dict):
    return Breakable(
        start="{",
        body=join(
            [Sequence(self(k), ": ", self(v)) for k, v in d.items()], sep=", "
        ),
        end="}",
    )


@ovld
def js_embed(self, seq: Union[list, tuple]):
    return Breakable(
        start="[",
        body=join(map(self, seq), sep=", "),
        end="]",
    )


@ovld
def js_embed(self, x: Union[int, float, str, bool, _type(None)]):
    return Text(json.dumps(x))


@ovld
def js_embed(self, expr: JSExpression):
    return expr.code


@ovld
def js_embed(self, t: HType.self):
    return "self"


@ovld
def js_embed(self, t: Tag):
    tag_id = t.get_attribute("id", None)
    if not tag_id:
        raise ValueError(f"Cannot embed <{t.name}> element without an id.")
    return f"document.getElementById('{tag_id}')"


@ovld
def js_embed(self, res: Resource):
    return self(res.obj)


@ovld
def js_embed(self, x: object):
    raise TypeError(
        f"Resources of type {type(x).__name__} cannot be serialized to JavaScript."
    )



import re
from html import escape
from types import SimpleNamespace
from typing import Sequence, Union

from ovld import OvldMC, ovld

from . import h as hmodule
from . import resource
from .embed import attr_embed, js_embed
from .h import H, Tag, iterate_children
from .textgen import Breakable, Text, TextFormatter


class ResourceDeduplicator:
    def __init__(self):
        self.seen_resources = set()

    def __call__(self, resource):
        if resource in self.seen_resources:
            return False
        else:
            self.seen_resources.add(resource)
            return True


def no_resources(resource):  # pragma: no cover
    return False


class HTMLGenerator(metaclass=OvldMC):
    def __init__(self, rules=None, *, attr_embed, js_embed):
        self.rules = rules or {}
        self.attr_embed = attr_embed
        self._js_embed = js_embed

    def fork(self, *, attr_embed=None, js_embed=None):
        return type(self)(
            dict(self.rules),
            attr_embed=attr_embed or self.attr_embed,
            js_embed=js_embed or self.js_embed,
        )

    def js_embed(self, obj, **fmt):
        x = self._js_embed(obj)
        if not isinstance(x, str):
            x = x.to_string(**fmt)
        return x

    def register(self, rule, fn=None):
        def reg(fn):
            self.rules[rule] = fn
            return fn

        if fn is None:
            return reg
        else:
            return reg(fn)

    @staticmethod
    def _default_tag(self, node, workspace, default):
        return default(node, workspace)

    @staticmethod
    def _default_attr(self, node, workspace, attr, value, default):
        return default(node, workspace, attr, value)

    def default_tag(self, node, workspace):
        workspace.open = node.name
        workspace.close = node.name

    def default_attr(self, node, workspace, attr, value):
        new_value = self.attr_embed(attr, value)
        if new_value is True:
            workspace.attributes[attr] = new_value
        elif isinstance(new_value, resource.JSExpression):
            workspace.attributes[attr] = self.expand_resources(
                new_value.code, self.js_embed
            )
        elif new_value is not None:
            workspace.attributes[attr] = self.expand_resources(
                new_value, self.attr_embed
            )

    @ovld
    def process(self, node: Union[str, TextFormatter]):
        return node

    @ovld
    def process(self, node: object):
        return str(node)

    @ovld
    def process(self, node: Tag):
        workspace = SimpleNamespace(
            open=None,
            close=None,
            attributes={},
            children=[],
            resources=list(node.resources),
            extra=[],
            escape_children=True,
        )

        workspace.children = [
            self.process(child) for child in iterate_children(node.children)
        ]
        for child in workspace.children:
            if getattr(child, "extra", None):
                workspace.extra += child.extra
                child.extra = []
            if getattr(child, "resources", None):
                workspace.resources += child.resources
                child.resources = []

        tag_rule = self.rules.get(f"tag:{node.name}", self._default_tag)
        if (
            tag_rule(self, node, workspace, self.default_tag) is False
        ):  # pragma: no cover
            pass
        else:
            for k, v in node.attributes.items():
                rule = self.rules.get(f"attr:{k}", self._default_attr)
                if (
                    rule(self, node, workspace, k, v, self.default_attr)
                    is False
                ):  # pragma: no cover
                    break

        return workspace

    def _text_parts(self, workspace):
        def convert_child(c):
            if isinstance(c, TextFormatter):
                return c
            elif isinstance(c, str):
                if workspace.escape_children:
                    return escape(str(c))
                else:
                    return c
            else:
                return self._text_parts(c)

        attr = "".join(
            f" {k}" if v is True else f' {k}="{escape(v)}"'
            for k, v in workspace.attributes.items()
        )

        children = list(map(convert_child, workspace.children))

        if workspace.open:
            if workspace.close:
                return Breakable(
                    start=f"<{workspace.open}{attr}>",
                    body=children,
                    end=f"</{workspace.open}>",
                )
            else:
                assert not children
                return Text(f"<{workspace.open}{attr} />")
        else:
            return Breakable(start=None, body=children, end=None)

    def expand_resources(self, value, embed):
        def sub(m):
            res = resource.registry.resolve(int(m.groups()[0]))
            return self.expand_resources(embed(res.obj), embed)

        return re.sub(
            pattern=rf"\[{resource.embed_key}:([0-9]+)\]",
            string=value,
            repl=sub,
        )

    def generate(self, node, filter_resources=True):
        ws = self.process(node)
        body = self._text_parts(ws)

        extra = Breakable(
            start=None,
            body=[self._text_parts(self.process(x)) for x in ws.extra],
            end=None,
        )

        if filter_resources is True:
            filter_resources = ResourceDeduplicator()
        elif filter_resources is None:  # pragma: no cover
            filter_resources = no_resources

        resources = []
        for r in ws.resources:
            if not filter_resources(r):
                continue
            entry, more_extra, more_resources = self.generate(r)
            assert not more_extra
            assert not more_resources
            resources.append(entry)

        resources = Breakable(start=None, body=resources, end=None)

        return body, extra, resources

    def __call__(self, node):
        body, extra, _ = self.generate(node)
        return f"{body}{extra}"


standard_html = HTMLGenerator(
    rules={}, attr_embed=attr_embed, js_embed=js_embed
)


######################################
# Handlers for void/raw/special tags #
######################################


def raw_tag(self, node, workspace, default):
    workspace.escape_children = False
    return default(node, workspace)


def void_tag(self, node, workspace, default):
    assert not workspace.children
    workspace.open = node.name
    workspace.close = None


def virtual_tag(self, node, workspace, default):
    return


def raw_virtual_tag(self, node, workspace, default):
    workspace.escape_children = False


# These tags are self-closing and cannot have children.
_void_tags = {
    "area",
    "base",
    "br",
    "col",
    "command",
    "embed",
    "hr",
    "img",
    "input",
    "keygen",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
}

for t in _void_tags:
    standard_html.register(f"tag:{t}", void_tag)


# We do not escape string children for these tags.
_raw_tags = {"style"}

for t in _raw_tags:
    standard_html.register(f"tag:{t}", raw_tag)


# These tags do not correspond to real HTML tags. They cannot have
# attributes, and their children are simply concatenated and inlined
# in the parent.
_virtual_tags = {None, "inline"}

for t in _virtual_tags:
    standard_html.register(f"tag:{t}", virtual_tag)


_raw_virtual_tags = {"raw"}

for t in _raw_virtual_tags:
    standard_html.register(f"tag:{t}", raw_virtual_tag)


###################################
# Handlers for special attributes #
###################################


@standard_html.register("tag:script")
def script_tag(self, node, workspace, default):
    workspace.escape_children = False
    new_children = []
    for child in workspace.children:
        assert isinstance(child, str)
        new_children.append(self.expand_resources(child, self.js_embed))
    workspace.children = new_children
    return default(node, workspace)


constructor_lib = H.script(
    """
$$HREPR = {
    prepare(node_id) {
        const self = document.getElementById(node_id);
        let resolve = null;
        self.__object = new Promise((rs, rj) => { resolve = rs });
        self.__object.__resolve = resolve;
    },
    isFunc(x) {
        let hasprop = prop => Object.getOwnPropertyNames(x).includes(prop);
        return (hasprop("arguments") || !hasprop("prototype"));
    }
}
"""
)


constructor_promise_script = "$$HREPR.prepare('{node_id}')"


constructor_template_script = """{imp}
const self = document.getElementById('{node_id}');
const arglist = {arguments};
let obj = $$HREPR.isFunc({symbol}) ? {symbol}(...arglist) : new {symbol}(...arglist);
self.__object.__resolve(obj);
"""


@standard_html.register("attr:--constructor")
def constructor_attribute(self, node, workspace, key, value, default):
    if "id" in node.attributes:
        node_id = node.attributes["id"]
    else:  # pragma: no cover
        # Setting __constructor should normally set an id, so this may not trigger
        autoid = f"$hrepr${next(hmodule.current_autoid)}"
        node_id = workspace.attributes["id"] = autoid

    module = value.get("module", None)
    script = value.get("script", None)

    workspace.resources.append(constructor_lib)
    if module:
        assert not isinstance(module, (list, tuple))
        assert script is None
        symbol = value.get("symbol", None)
        module = self.js_embed(module)
        if symbol is None:
            symbol = "constructor"
            imp = f"import constructor from {module};"
        elif symbol.startswith("default."):
            imp = f"import dflt from {module};"
            symbol = symbol.replace("default.", "dflt.")
        else:
            rootsym, sep, props = symbol.partition(".")
            imp = f"import {{{rootsym}}} from {module};"
    else:
        if script:
            script = [
                src
                if isinstance(src, Tag)
                else H.script(src=src, type="text/javascript")
                for src in (
                    script if isinstance(script, (list, tuple)) else [script]
                )
            ]
            workspace.resources.extend(script)
        symbol = value["symbol"]
        imp = ""

    if "options" in value:
        assert "arguments" not in value
        arguments = [H.self(), value["options"]]
    elif "arguments" in value:
        arguments = value["arguments"]
        if not isinstance(arguments, (list, tuple)):
            arguments = [arguments]
    else:
        arguments = [H.self()]

    if "stylesheet" in value:
        sheet = value["stylesheet"]
        sheet = [
            src if isinstance(src, Tag) else H.link(rel="stylesheet", href=src)
            for src in (sheet if isinstance(sheet, (list, tuple)) else [sheet])
        ]
        workspace.resources.extend(sheet)

    sc_promise = H.script(constructor_promise_script.format(node_id=node_id))
    workspace.extra.append(sc_promise)

    sc = H.script(
        constructor_template_script.format(
            imp=imp,
            symbol=symbol,
            node_id=node_id,
            arguments=self.js_embed(arguments),
        ),
        type="module",
    )

    workspace.extra.append(sc)


@standard_html.register("attr:--resources")
def attr_resources(self, node, workspace, key, value, default):
    if not isinstance(value, Sequence):
        workspace.resources.append(value)
    else:
        workspace.resources.extend(value)


@standard_html.register("attr:style")
def attr_style(self, node, workspace, key, value, default):
    if isinstance(value, dict):
        value = "".join(f"{k}:{v};" for k, v in value.items())
    workspace.attributes["style"] = value



"""Extensible HTML representation for Python objects."""

from . import h
from .core import Config, Hrepr, HreprState, Interface, StdHrepr  # noqa: F401
from .embed import attr_embed, js_embed  # noqa: F401
from .h import HTML, H, HType, Tag  # noqa: F401
from .hgen import standard_html

h.standard_html = standard_html

config_defaults = {
    "string_cutoff": 20,
    "bytes_cutoff": 20,
    "sequence_max": 100,
}

hrepr = Interface(StdHrepr, **config_defaults)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hrepr']

package_data = \
{'': ['*'], 'hrepr': ['style/*']}

install_requires = \
['ovld>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'hrepr',
    'version': '0.5.1',
    'description': 'Extensible HTML representation for Python objects.',
    'long_description': '# hrepr\n\n`hrepr` outputs HTML/pretty representations for Python objects.\n\n✅ Nice, colourful representations of lists, dicts, dataclasses, booleans...<br/>\n✅ Ridiculously extensible and configurable<br/>\n✅ Handles recursive data structures<br/>\n✅ Compatible with Jupyter notebooks<br/>\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr1.png" width="400px"><img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr2.png" width="400px">\n\nI suggest studying the example file to learn `hrepr`:\n\n* `python examples/exhibit.py > exhibit.html` (and then view the HTML file)\n\nAlso see the Jupyter notebook at `examples/Basics.ipynb`, but keep in mind that GitHub can\'t display it properly because of the injected styles/scripts.\n\n\n## Install\n\n```python\npip install hrepr\n```\n\n\n## Usage\n\n```python\nfrom hrepr import hrepr\nobj = {\'potatoes\': [1, 2, 3], \'bananas\': {\'cantaloups\': 8}}\n\n# Print the HTML representation of obj\nprint(hrepr(obj))\n\n# Wrap the representation in <html><body> tags and embed the default\n# css style files in a standalone page, which is saved to obj.html\nhrepr.page(obj, file="obj.html")\n```\n\nIn a Jupyter Notebook, you can return `hrepr(obj)` from any cell and it will show its representation for you. You can also write `display_html(hrepr(obj))`.\n\n\n## Custom representations\n\nA custom representation for an object can be defined using the following three methods (it is not necessary to define all of them, only those that are relevant to your case):\n\n* `__hrepr__(self, H, hrepr)` returns the normal HTML representation.\n    * Use `H.span["some-class"](some-content, some_attr=some_value)` to generate HTML.\n    * Use `hrepr(self.x)` to generate the representation for some subfield `x`.\n    * `hrepr.config` contains any keyword arguments given in the top level call to `hrepr`. For instance, if you call `hrepr(obj, blah=3)`, then `hrepr.config.blah == 3` in all calls to `__hrepr__` down the line (the default value for all keys is `None`).\n* `__hrepr_short__(self, H, hrepr)` returns a *short* representation, ideally of a constant size.\n    * The output of this method is used when we hit max depth, or for repeated references.\n    * Only include bare minimum information. Short means short.\n* `__hrepr_resources__(cls, H)` is a **classmethod** that returns resources common to all instances of the class (typically a stylesheet or a script).\n    * When generating a page, the resources will go in `<head>`.\n    * You can return a list of resources.\n\nNo dependency on `hrepr` is necessary.\n\nFor example:\n\n```python\nclass Person:\n    def __init__(self, name, age, job):\n        self.name = name\n        self.age = age\n        self.job = job\n\n    @classmethod\n    def __hrepr_resources__(cls, H):\n        # Note: you might need to add "!important" next to some rules if\n        # they conflict with defaults from hrepr\'s own CSS.\n        return H.style("""\n            .person {\n                background: magenta !important;\n                border-color: magenta !important;\n            }\n            .person-short { font-weight: bold; color: green; }\n        """)\n\n    def __hrepr__(self, H, hrepr):\n        # hrepr.make.instance is a helper to show a table with a header that\n        # describes some kind of object\n        return hrepr.make.instance(\n            title=self.name,\n            fields=[["age", self.age], ["job", self.job]],\n            delimiter=" ↦ ",\n            type="person",\n        )\n\n    def __hrepr_short__(self, H, hrepr):\n        return H.span["person-short"](self.name)\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr3.png" width="600px">\n\n\n## References\n\n`hrepr` can handle circular references. Furthermore, if an object is found at several places in a structure, only the first occurrence will be printed in full, and any other will be a numeric reference mapped to the short representation for the object. It looks like this:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr4.png" width="600px">\n\nThe `shortrefs` and `norefs` configuration keys control the representation of references:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr5.png" width="600px">\n\n`norefs` is ignored when there are circular references.\n\n\n## HTML generation\n\nGenerate HTML using the `H` parameter to `__hrepr__`, or import it and use it directly:\n\n```python\nfrom hrepr import H\nhtml = H.span["bear"](\n    "Only ", H.b("YOU"), " can prevent forest fires!",\n    style="color: brown;"\n)\nprint(html)\n# <span class="bear" style="color: brown;">Only <b>YOU</b> can prevent forest fires!</span>\n```\n\n`H` can be built incrementally: if you have an element, you can call it to add children, index it to add classes, and so on. For instance:\n\n```python\nfrom hrepr import H\nhtml = H.span()\nhtml = html("Only ")\nhtml = html(style="color: brown;")["bear"]\nhtml = html(H.b("YOU"), " can prevent forest fires!")\nprint(html)\n# <span class="bear" style="color: brown;">Only <b>YOU</b> can prevent forest fires!</span>\n```\n\nThis can be handy if you want to tweak generated HTML a little. For example, `hrepr(obj)["fox"]` will tack on the class `fox` to the representation of the object.\n\n\n### Helpers\n\n* `hrepr.make.instance(title, fields, delimiter=None, type=None)`: formats the fields like a dataclass, with title on top.\n* `hrepr.make.bracketed(body, start, end, type=None)`: formats the body with the specified start/end bracket.\n\n\n### Constructed elements\n\nTo make it a bit easier to include and use JavaScript libraries, you can use the special `__constructor` attribute.\n\nFor example, you can load Plotly and create a plot like this:\n\n```python\ndef Plot(data):\n    return H.div(\n        __constructor={\n            "script": "https://cdn.plot.ly/plotly-latest.min.js",\n            "symbol": "Plotly.newPlot",\n            "options": [{"x": list(range(len(data))), "y": list(data)}],\n        }\n    )\nprint(Plot([math.sin(x / 10) for x in range(100)]))\n```\n\nThe above will:\n\n* Load the specified script.\n* Get the `Plotly.newPlot` function in the global namespace.\n* Call it with the `div` element as the first argument, and the `options` as the second argument.\n\nIt will look like this:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr6.png" width="600px">\n\n\n### Modules\n\nAnother example, this time using ESM (modules):\n\n```python\nnode = H.div(\n    style="width:500px;height:500px;border:1px solid black;",\n    __constructor={\n        "module": "https://cdnjs.cloudflare.com/ajax/libs/cytoscape/3.23.0/cytoscape.esm.min.js",\n        "arguments": {\n            "container": H.self(),\n            "elements": [\n                {"data": {"id": "A"}},\n                {"data": {"id": "B"}},\n                {"data": {"id": "C"}},\n                {"data": {"source": "A", "target": "B"}},\n                {"data": {"source": "B", "target": "C"}},\n                {"data": {"source": "C", "target": "A"}},\n            ],\n            "style": cystyle,\n            "layout": {"name": "cose"},\n        },\n    },\n)\nprint(node)\n```\n\nThe above will:\n\n* Import the specified module.\n* Call the module\'s default export with `arguments`.\n  * Note the use of `H.self()` to refer to the target `div` in the arguments.\n\nIf you wish to use a non-default export, set the `symbol` key in the `__constructor` attribute to the name of the export you want.\n\n\n## Customize hrepr\n\n### Mixins\n\nIf you want to *really* customize hrepr, you can use mixins. They look like a bit of black magic, but they\'re simple enough:\n\n```python\n# ovld is one of the dependencies of hrepr\nfrom ovld import ovld, extend_super, has_attribute, OvldMC\nfrom hrepr import hrepr\n\nclass MyMixin(metaclass=OvldMC):\n    # Change the representation of integers\n\n    @extend_super\n    def hrepr_resources(self, cls: int):\n        # Note: in hrepr_resources, cls is the int type, not an integer\n        return self.H.style(".my-integer { color: fuchsia; }")\n\n    @extend_super\n    def hrepr(self, n: int):\n        return self.H.span["my-integer"]("The number ", str(n))\n\n    # Specially handle any object with a "quack" method\n\n    def hrepr(self, duck: has_attribute("quack")):\n        return self.H.span("🦆")\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr7.png" width="600px">\n\nThe annotation for a rule can either be a type, `ovld.has_attribute`, or pretty much any function wrapped with the `ovld.meta` decorator, as long as the function operates on classes. See the documentation for [ovld](https://github.com/breuleux/ovld#other-features) for more information.\n\nAnd yes, you can define `hrepr` multiple times inside the class, as long as they have distinct annotations and you inherit from `Hrepr`. You can also define `hrepr_short` or `hrepr_resources` the same way.\n\n### Postprocessors\n\n`hrepr` can be given a postprocessor that is called on the representation of any object. You can use this to do things like highlighting specific objects:\n\n```python\nfrom hrepr import H\n\nstyle = H.style(".highlight { border: 3px solid red !important; }")\n\ndef highlight(x):\n    def postprocess(element, obj, hrepr):\n        if obj == x:\n            # Adds the "highlight" class and attaches a style\n            return element["highlight"].fill(resources=style)\n        else:\n            return element\n\n    return postprocess\n\nhrepr([1, 2, [3, 4, 2]], postprocess=highlight(2))\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr8.png" width="600px">\n\n\n### hrepr variants\n\nTo put this all together, you can create a *variant* of `hrepr`:\n\n```python\nhrepr2 = hrepr.variant(mixins=MyMixin, postprocess=highlight(2))\nhrepr2([1, 2, 3])  # Will use the mixins and postprocessor\n```\n\n\n### Configure the hrepr function itself\n\nAlternatively, you can configure the main `hrepr`:\n\n```python\nhrepr.configure(mixins=MyMixin, postprocess=highlight(2))\n```\n\nBut keep in mind that unlike the variant, the above will modify `hrepr` for everything else as well.\n',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/breuleux/hrepr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
